using System.Collections.Generic;

namespace Syntriax.Engine.Core;

/// <summary>
/// Represents a collector for the class type of <typeparamref name="T"/>.
/// Provides mechanisms for tracking additions and removals, and notifies subscribers when such events occur on the assigned <see cref="IUniverse"/>.
/// </summary>
/// <typeparam name="T">The type of objects tracked by the collector.</typeparam>
public interface IBehaviourCollector<T> : IHasUniverse, IEnumerable<T> where T : class
{
    /// <summary>
    /// Event triggered when an object of type <typeparamref name="T"/> is added to the collector.
    /// </summary>
    event CollectedEventHandler? OnCollected;

    /// <summary>
    /// Event triggered when an object of type <typeparamref name="T"/> is removed from the collector.
    /// </summary>
    event RemovedEventHandler? OnRemoved;

    /// <summary>
    /// Delegate for handling the <see cref="OnCollected"/> event.
    /// </summary>
    /// <param name="sender">The instance of the <see cref="IBehaviourCollector{T}"/> that triggered the event.</param>
    /// <param name="behaviourCollected">The object of type <typeparamref name="T"/> that was added to the collector.</param>
    delegate void CollectedEventHandler(IBehaviourCollector<T> sender, T behaviourCollected);

    /// <summary>
    /// Delegate for handling the <see cref="OnRemoved"/> event.
    /// </summary>
    /// <param name="sender">The instance of the <see cref="IBehaviourCollector{T}"/> that triggered the event.</param>
    /// <param name="behaviourRemoved">The object of type <typeparamref name="T"/> that was removed from the collector.</param>
    delegate void RemovedEventHandler(IBehaviourCollector<T> sender, T behaviourRemoved);
}
