using System.Collections.Generic;

namespace Syntriax.Engine.Core;

/// <summary>
/// Represents a universe responsible for managing <see cref="IUniverseObject"/>s.
/// </summary>
public interface IUniverse : IEntity, IEnumerable<IUniverseObject>
{
    /// <summary>
    /// Event triggered when <see cref="Update(UniverseTime)"/> is about to be called called on the <see cref="IUniverse"/>.
    /// </summary>
    event UpdateEventHandler? OnPreUpdate;

    /// <summary>
    /// Event triggered when <see cref="Update(UniverseTime)"/> is called on the <see cref="IUniverse"/>.
    /// </summary>
    event UpdateEventHandler? OnUpdate;

    /// <summary>
    /// Event triggered when <see cref="PreDraw"/> is called on the <see cref="IUniverse"/>.
    /// </summary>
    event PreDrawEventHandler? OnPreDraw;

    /// <summary>
    /// Event triggered when a <see cref="IUniverseObject"/> is registered to the <see cref="IUniverse"/>.
    /// </summary>
    event UniverseObjectRegisteredEventHandler? OnUniverseObjectRegistered;

    /// <summary>
    /// Event triggered when a <see cref="IUniverseObject"/> is unregistered from the <see cref="IUniverse"/>.
    /// </summary>
    event UniverseObjectUnRegisteredEventHandler? OnUniverseObjectUnRegistered;

    /// <summary>
    /// Event triggered when <see cref="TimeScale"/> is changed on the <see cref="IUniverse"/>.
    /// </summary>
    event TimeScaleChangedEventHandler? OnTimeScaleChanged;

    /// <summary>
    /// Current time scale the <see cref="IUniverse"/> operates on.
    /// </summary>
    float TimeScale { get; set; }

    /// <summary>
    /// Contains time data related to this <see cref="IUniverse"/>.
    /// </summary>
    UniverseTime Time { get; }

    /// <summary>
    /// Contains unscaled time data related to this <see cref="IUniverse"/>.
    /// </summary>
    UniverseTime UnscaledTime { get; }

    /// <summary>
    /// Gets a read-only list of <see cref="IUniverseObject"/>s managed by the <see cref="IUniverse"/>.
    /// </summary>
    IReadOnlyList<IUniverseObject> UniverseObjects { get; }

    /// <summary>
    /// Registers an <see cref="IUniverseObject"/> to the <see cref="IUniverse"/>.
    /// </summary>
    /// <param name="universeObject">The <see cref="IUniverseObject"/> to register.</param>
    void Register(IUniverseObject universeObject);

    /// <summary>
    /// Instantiates a <see cref="IUniverseObject"/> of type T with the given arguments and registers it to the <see cref="IUniverse"/>.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IUniverseObject"/> to instantiate.</typeparam>
    /// <param name="args">Constructor parameters for the given type of <see cref="IUniverseObject"/>.</param>
    /// <returns>The instantiated <see cref="IUniverseObject"/>.</returns>
    T InstantiateUniverseObject<T>(params object?[]? args) where T : class, IUniverseObject;

    /// <summary>
    /// Removes an <see cref="IUniverseObject"/> from the <see cref="IUniverse"/>.
    /// </summary>
    /// <param name="universeObject">The <see cref="IUniverseObject"/> to remove.</param>
    void Remove(IUniverseObject universeObject);

    /// <summary>
    /// Updates the <see cref="IUniverse"/> with the given delta time.
    /// </summary>
    /// <param name="universeTime">Delta time.</param>
    void Update(UniverseTime universeTime);

    /// <summary>
    /// Performs operations that should be done before the draw calls.
    /// </summary>
    void PreDraw();

    delegate void TimeScaleChangedEventHandler(IUniverse sender, float previousTimeScale);

    delegate void UpdateEventHandler(IUniverse sender, UniverseTime engineTime);
    delegate void PreDrawEventHandler(IUniverse sender);

    delegate void UniverseObjectRegisteredEventHandler(IUniverse sender, IUniverseObject universeObjectRegistered);
    delegate void UniverseObjectUnRegisteredEventHandler(IUniverse sender, IUniverseObject universeObjectUnregistered);
}
