using System.Collections.Generic;

namespace Syntriax.Engine.Core;

/// <summary>
/// Represents an <see cref="IEntity"/> that can enter and exit a universe within the <see cref="IUniverse"/> system.
/// This interface allows for tracking the object's presence in the universe and provides events 
/// for notifying when the see enters or exits the universe.
/// </summary>
public interface IUniverseObject : IEntity, IActive, INameable, IHasBehaviourController, IEnumerable<IUniverseObject>
{
    /// <summary>
    /// Event triggered when the <see cref="IUniverseObject"/> enters the universe.
    /// </summary>
    event EnteredUniverseEventHandler? OnEnteredUniverse;

    /// <summary>
    /// Event triggered when the <see cref="IUniverseObject"/> exits the universe.
    /// </summary>
    event ExitedUniverseEventHandler? OnExitedUniverse;

    /// <summary>
    /// Event triggered when the <see cref="Parent"/> of the <see cref="IUniverseObject"/> changes. The second parameter is the old <see cref="IUniverseObject"/>.
    /// </summary>
    event ParentChangedEventHandler? OnParentChanged;

    /// <summary>
    /// Event triggered when a new <see cref="IUniverseObject"/> is added to the <see cref="Children"/>.
    /// </summary>
    event ChildrenAddedEventHandler? OnChildrenAdded;

    /// <summary>
    /// Event triggered when an <see cref="IUniverseObject"/> is removed from the <see cref="Children"/>.
    /// </summary>
    event ChildrenRemovedEventHandler? OnChildrenRemoved;

    /// <summary>
    /// Gets the <see cref="IUniverse"/> this <see cref="IUniverseObject"/> is connected to, if any.
    /// </summary>
    IUniverse Universe { get; }

    /// <summary>
    /// Indicates whether the <see cref="IUniverseObject"/> is currently in the universe.
    /// </summary>
    bool IsInUniverse { get; }

    /// <summary>
    /// The parent <see cref="IUniverseObject"/> of the <see cref="IUniverseObject"/>.
    /// </summary>
    IUniverseObject? Parent { get; }

    /// <summary>
    /// The <see cref="IUniverseObject"/>s that have this <see cref="IUniverseObject"/> as their <see cref="Parent"/>.
    /// </summary>
    IReadOnlyList<IUniverseObject> Children { get; }

    /// <summary>
    /// Internal method to handle entering the universe.
    /// This should be called by the system to properly manage universe states.
    /// </summary>
    /// <param name="universe">The <see cref="IUniverse"/> that is managing this universe.</param>
    /// <returns>
    /// <see cref="true"/> if the <see cref="IUniverseObject"/> successfully entered the universe; 
    /// <see cref="false"/> if it failed to do so.
    /// </returns>
    internal bool EnterUniverse(IUniverse universe);

    /// <summary>
    /// Internal method to handle exiting the universe.
    /// This should be called by the system to properly manage universe states.
    /// </summary>
    /// <returns>
    /// <see cref="true"/> if the <see cref="IUniverseObject"/> successfully exited the universe; 
    /// <see cref="false"/> if it failed to do so.
    /// </returns>
    internal bool ExitUniverse();

    /// <summary>
    /// Sets the parent <see cref="IUniverseObject"/> of this <see cref="IUniverseObject"/>.
    /// </summary>
    /// <param name="universeObject">The parent <see cref="IUniverseObject"/> to set.</param>
    void SetParent(IUniverseObject? universeObject);

    /// <summary>
    /// Adds a child <see cref="IUniverseObject"/> to this <see cref="IUniverseObject"/>.
    /// </summary>
    /// <param name="universeObject">The child <see cref="IUniverseObject"/> to add.</param>
    void AddChild(IUniverseObject universeObject);

    /// <summary>
    /// Removes a child <see cref="IUniverseObject"/> from this <see cref="IUniverseObject"/>.
    /// </summary>
    /// <param name="universeObject">The child <see cref="IUniverseObject"/> to remove.</param>
    void RemoveChild(IUniverseObject universeObject);

    /// <summary>
    /// EventHandler delegate for the event triggered when the <see cref="IUniverseObject"/> enters the universe of a <see cref="IUniverse">.
    /// </summary>
    /// <param name="sender">The <see cref="IUniverseObject"/> that entered the universe.</param>
    /// <param name="universe">The <see cref="IUniverse"/> that the <see cref="IUniverseObject"/> has entered it's universe.</param>
    delegate void EnteredUniverseEventHandler(IUniverseObject sender, IUniverse universe);

    /// <summary>
    /// EventHandler delegate for the event triggered when the <see cref="IUniverseObject"/> exits the universe of a <see cref="IUniverse">.
    /// </summary>
    /// <param name="sender">The <see cref="IUniverseObject"/> that exited the universe.</param>
    /// <param name="universe">The <see cref="IUniverse"/> that the <see cref="IUniverseObject"/> has exited it's universe.</param>
    delegate void ExitedUniverseEventHandler(IUniverseObject sender, IUniverse universe);

    /// <summary>
    /// Delegate for the event triggered when the <see cref="IUniverseObject"/>'s parent changes.
    /// </summary>
    /// <param name="sender">The <see cref="IUniverseObject"/> that the parent has changed.</param>
    /// <param name="previousParent">The previous <see cref="IUniverseObject"/> the sender was a child of.</param>
    /// <param name="newParent">The new and current <see cref="IUniverseObject"/> the sender is a child of.</param>
    delegate void ParentChangedEventHandler(IUniverseObject sender, IUniverseObject? previousParent, IUniverseObject? newParent);

    /// <summary>
    /// Delegate for the event triggered when a new <see cref="IUniverseObject"/> added as a child.
    /// </summary>
    /// <param name="sender">The parent <see cref="IUniverseObject"/> this event is being called from.</param>
    /// <param name="childrenAdded">The <see cref="IUniverseObject"/> that got removed as a children of the sender <see cref="IUniverseObject"/>.</param>
    delegate void ChildrenAddedEventHandler(IUniverseObject sender, IUniverseObject childrenAdded);

    /// <summary>
    /// Delegate for the event triggered when a new <see cref="IUniverseObject"/> removed from being a child.
    /// </summary>
    /// <param name="sender">The parent <see cref="IUniverseObject"/> this event is being called from.</param>
    /// <param name="childrenAdded">The <see cref="IUniverseObject"/> that got removed as a children of the sender <see cref="IUniverseObject"/>.</param>
    delegate void ChildrenRemovedEventHandler(IUniverseObject sender, IUniverseObject childrenRemoved);
}
