using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Syntriax.Engine.Core;

[System.Diagnostics.DebuggerDisplay("Behaviour Count: {behaviours.Count}")]
public class BehaviourController : BaseEntity, IBehaviourController
{
    public event IBehaviourController.PreUpdateEventHandler? OnPreUpdate = null;
    public event IBehaviourController.UpdateEventHandler? OnUpdate = null;
    public event IBehaviourController.PreDrawEventHandler? OnPreDraw = null;

    public event IBehaviourController.BehaviourAddedEventHandler? OnBehaviourAdded = null;
    public event IBehaviourController.BehaviourRemovedEventHandler? OnBehaviourRemoved = null;
    public event IHasUniverseObject.UniverseObjectAssignedEventHandler? OnUniverseObjectAssigned = null;

    private readonly IList<IBehaviour> behaviours = new List<IBehaviour>(Constants.BEHAVIOURS_SIZE_INITIAL);

    private IUniverseObject _universeObject = null!;

    public IUniverseObject UniverseObject => _universeObject;

    public T AddBehaviour<T>(T behaviour) where T : class, IBehaviour
    {
        InsertBehaviourByPriority(behaviour);

        behaviour.Assign(this);

        if (IsInitialized)
            behaviour.Initialize();
        behaviour.OnPriorityChanged += OnPriorityChange;
        OnBehaviourAdded?.InvokeSafe(this, behaviour);
        return behaviour;
    }

    public T AddBehaviour<T>(params object?[]? args) where T : class, IBehaviour
        => AddBehaviour(Factory.BehaviourFactory.Instantiate<T>(_universeObject, args));

    public T? GetBehaviour<T>()
    {
        foreach (IBehaviour behaviourItem in behaviours)
            if (behaviourItem is T result)
                return result;

        return default;
    }

    public IReadOnlyList<T> GetBehaviours<T>()
    {
        List<T>? behaviours = null;
        foreach (IBehaviour behaviourItem in this.behaviours)
        {
            if (behaviourItem is not T behaviour)
                continue;

            behaviours ??= [];
            behaviours.Add(behaviour);
        }

        return behaviours ?? Enumerable.Empty<T>().ToList();
    }

    public void GetBehaviours<T>(IList<T> results)
    {
        results.Clear();
        foreach (IBehaviour behaviourItem in behaviours)
        {
            if (behaviourItem is not T behaviour)
                continue;

            results.Add(behaviour);
        }
    }

    public void RemoveBehaviour<T>(bool removeAll = false) where T : class, IBehaviour
    {
        for (int i = behaviours.Count; i >= 0; i--)
        {
            if (behaviours[i] is not T behaviour)
                continue;

            RemoveBehaviour(behaviour);

            if (!removeAll)
                return;
        }
    }

    public void RemoveBehaviour<T>(T behaviour) where T : class, IBehaviour
    {
        if (!behaviours.Contains(behaviour))
            throw new Exception($"{behaviour.GetType().Name} does not exist in {UniverseObject.Name}'s {nameof(IBehaviourController)}.");

        behaviour.OnPriorityChanged -= OnPriorityChange;
        behaviour.Finalize();
        behaviours.Remove(behaviour);
        OnBehaviourRemoved?.InvokeSafe(this, behaviour);
    }

    protected virtual void OnAssign(IUniverseObject universeObject) { }
    public bool Assign(IUniverseObject universeObject)
    {
        if (UniverseObject is not null && UniverseObject.IsInitialized)
            return false;

        _universeObject = universeObject;
        OnAssign(universeObject);
        OnUniverseObjectAssigned?.InvokeSafe(this);
        return true;
    }

    protected override void InitializeInternal()
    {
        Debug.Assert.AssertUniverseObjectAssigned(this);

        foreach (IBehaviour behaviour in behaviours)
            behaviour.Initialize();
    }

    protected override void FinalizeInternal()
    {
        foreach (IBehaviour behaviour in behaviours)
            behaviour.Finalize();
    }

    public void Update()
    {
        Debug.Assert.AssertInitialized(this);

        if (!UniverseObject.StateEnable.Enabled || !StateEnable.Enabled)
            return;

        OnPreUpdate?.InvokeSafe(this);
        OnUpdate?.InvokeSafe(this);
    }

    public void UpdatePreDraw()
    {
        Debug.Assert.AssertInitialized(this);

        if (!UniverseObject.StateEnable.Enabled || !StateEnable.Enabled)
            return;

        OnPreDraw?.InvokeSafe(this);
    }

    public BehaviourController() { }
    public BehaviourController(IUniverseObject universeObject) => Assign(universeObject);

    private void InsertBehaviourByPriority<T>(T behaviour) where T : class, IBehaviour
    {
        int i;

        for (i = 0; i < behaviours.Count; i++)
        {
            if (behaviours[i].Priority > behaviour.Priority)
                continue;

            behaviours.Insert(i, behaviour);
            return;
        }

        if (i == 0 || i == behaviours.Count)
            behaviours.Add(behaviour);
    }

    private void OnPriorityChange(IBehaviour sender, int previousPriority)
    {
        behaviours.Remove(sender);
        InsertBehaviourByPriority(sender);
    }

    public IEnumerator<IBehaviour> GetEnumerator() => behaviours.GetEnumerator();
    IEnumerator IEnumerable.GetEnumerator() => behaviours.GetEnumerator();
}
