using System.Diagnostics.CodeAnalysis;

using Syntriax.Engine.Core.Exceptions;

namespace Syntriax.Engine.Core;

public static class BehaviourControllerExtensions
{
    /// <summary>
    /// Tries to get a <see cref="IBehaviour"/> of the specified type.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to get.</typeparam>
    /// <param name="behaviourController">The <see cref="IBehaviourController"/> to search in.</param>
    /// <param name="behaviour">When this method returns, contains the <see cref="IBehaviour"/> of the specified type, if found; otherwise, null.</param>
    /// <returns><see cref="true"/> if a <see cref="IBehaviour"/> of the specified type was found; otherwise, <see cref="false"/>.</returns>
    public static bool TryGetBehaviour<T>(this IBehaviourController behaviourController, [NotNullWhen(returnValue: true)] out T? behaviour)
    {
        behaviour = behaviourController.GetBehaviour<T>();
        return behaviour is not null;
    }

    /// <summary>
    /// Gets a <see cref="IBehaviour"/> of the specified type in the provided <see cref="IBehaviourController"/>. Throws an error if not found.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to get.</typeparam>
    /// <param name="behaviourController">The <see cref="IBehaviourController"/> to start searching from.</param>
    /// <returns>The <see cref="IBehaviour"/> of the specified type if found; otherwise, throws <see cref="BehaviourNotFoundException"/>.</returns>
    public static T GetRequiredBehaviour<T>(this IBehaviourController behaviourController) where T : class
        => behaviourController.GetBehaviour<T>() ?? throw new BehaviourNotFoundException($"{behaviourController.UniverseObject.Name}'s {nameof(IBehaviourController)} does not contain any {typeof(T).FullName}");

    /// <summary>
    /// Gets an existing <see cref="IBehaviour"/> of the specified type, or adds and returns a new one if it doesn't exist.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to get or add.</typeparam>
    /// <param name="behaviourController">The <see cref="IBehaviourController"/> to search in.</param>
    /// <param name="args">Optional arguments to pass to the constructor of the <see cref="IBehaviour"/> if a new one is added.</param>
    /// <returns>The existing or newly added <see cref="IBehaviour"/> of the specified type.</returns>
    public static T GetOrAddBehaviour<T>(this IBehaviourController behaviourController, params object?[]? args) where T : class, IBehaviour
        => behaviourController.GetBehaviour<T>() ?? behaviourController.AddBehaviour<T>(args);

    /// <summary>
    /// Tries to get a <see cref="IBehaviour"/> of the specified type in it's <see cref="IUniverseObject"/>'s parents recursively.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to get.</typeparam>
    /// <param name="behaviourController">The <see cref="IBehaviourController"/> to start searching from.</param>
    /// <param name="behaviour">When this method returns, contains the <see cref="IBehaviour"/> of the specified type, if found; otherwise, null.</param>
    /// <returns><see cref="true"/> if a <see cref="IBehaviour"/> of the specified type was found in the parent universe; otherwise, <see cref="false"/>.</returns>
    public static bool TryGetBehaviourInParent<T>(this IBehaviourController behaviourController, [NotNullWhen(returnValue: true)] out T? behaviour) where T : class
    {
        behaviour = GetBehaviourInParent<T>(behaviourController);
        return behaviour is not null;
    }

    /// <summary>
    /// Gets a <see cref="IBehaviour"/> of the specified type in it's <see cref="IUniverseObject"/>'s parents recursively.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to get.</typeparam>
    /// <param name="behaviourController">The <see cref="IBehaviourController"/> to start searching from.</param>
    /// <returns>The <see cref="IBehaviour"/> of the specified type if found; otherwise, null.</returns>
    public static T? GetBehaviourInParent<T>(this IBehaviourController behaviourController) where T : class
    {
        IBehaviourController? controller = behaviourController;

        while (controller is not null)
        {
            if (controller.GetBehaviour<T>() is T behaviour)
                return behaviour;

            controller = controller.UniverseObject.Parent?.BehaviourController;
        }

        return default;
    }

    /// <summary>
    /// Gets a <see cref="IBehaviour"/> of the specified type in it's <see cref="IUniverseObject"/>'s parents recursively. Throws an error if not found.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to get.</typeparam>
    /// <param name="behaviourController">The <see cref="IBehaviourController"/> to start searching from.</param>
    /// <returns>The <see cref="IBehaviour"/> of the specified type if found; otherwise, throws <see cref="BehaviourNotFoundException"/>.</returns>
    public static T GetRequiredBehaviourInParent<T>(this IBehaviourController behaviourController) where T : class
        => behaviourController.GetBehaviourInParent<T>() ?? throw new BehaviourNotFoundException($"{behaviourController.UniverseObject.Name}'s {nameof(IBehaviourController)} does not contain any {typeof(T).FullName} on any parent");

    /// <summary>
    /// Tries to get a <see cref="IBehaviour"/> of the specified type in it's <see cref="IUniverseObject"/>'s children recursively.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to get.</typeparam>
    /// <param name="behaviourController">The <see cref="IBehaviourController"/> to start searching from.</param>
    /// <param name="behaviour">When this method returns, contains the <see cref="IBehaviour"/> of the specified type, if found; otherwise, null.</param>
    /// <returns><see cref="true"/> if a <see cref="IBehaviour"/> of the specified type was found in the child universe; otherwise, <see cref="false"/>.</returns>
    public static bool TryGetBehaviourInChildren<T>(this IBehaviourController behaviourController, [NotNullWhen(returnValue: true)] out T? behaviour) where T : class
    {
        behaviour = GetBehaviourInChildren<T>(behaviourController);
        return behaviour is not null;
    }

    /// <summary>
    /// Gets a <see cref="IBehaviour"/> of the specified type in it's <see cref="IUniverseObject"/>'s children recursively.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to get.</typeparam>
    /// <param name="behaviourController">The <see cref="IBehaviourController"/> to start searching from.</param>
    /// <returns>The <see cref="IBehaviour"/> of the specified type if found; otherwise, null.</returns>
    public static T? GetBehaviourInChildren<T>(this IBehaviourController behaviourController) where T : class
    {
        if (behaviourController.GetBehaviour<T>() is T localBehaviour)
            return localBehaviour;

        foreach (IUniverseObject child in behaviourController.UniverseObject)
            if (GetBehaviourInChildren<T>(child.BehaviourController) is T behaviour)
                return behaviour;

        return default;
    }

    /// <summary>
    /// Gets a <see cref="IBehaviour"/> of the specified type in the children recursively. Throws an error if not found.
    /// </summary>
    /// <typeparam name="T">The type of <see cref="IBehaviour"/> to get.</typeparam>
    /// <param name="behaviourController">The <see cref="IBehaviourController"/> to start searching from.</param>
    /// <returns>The <see cref="IBehaviour"/> of the specified type if found; otherwise, throws <see cref="BehaviourNotFoundException"/>.</returns>
    public static T GetRequiredBehaviourInChildren<T>(this IBehaviourController behaviourController) where T : class
        => behaviourController.GetBehaviourInChildren<T>() ?? throw new BehaviourNotFoundException($"{behaviourController.UniverseObject.Name}'s {nameof(IBehaviourController)} does not contain any {typeof(T).FullName} on any children ");
}
