using System.Collections.Generic;

using Syntriax.Engine.Core;

namespace Syntriax.Engine.Systems.StateMachine;

public class State : BaseEntity, IState
{
    public event IState.StateUpdateEventHandler? OnStateUpdate = null;
    public event IState.StateTransitionedFromEventHandler? OnStateTransitionedFrom = null;
    public event IState.StateTransitionedToEventHandler? OnStateTransitionedTo = null;
    public event IState.StateTransitionReadyEventHandler? OnStateTransitionReady = null;
    public event INameable.NameChangedEventHandler? OnNameChanged = null;

    private readonly List<StateTransition> transitions = [];
    private readonly Dictionary<string, StateTransition> possibleTransitions = [];
    private string _name = "Default State Name";

    public IReadOnlyList<StateTransition> Transitions => transitions;
    public IReadOnlyDictionary<string, StateTransition> PossibleTransitions => possibleTransitions;
    public string Name
    {
        get => _name;
        set
        {
            if (_name.CompareTo(value) == 0)
                return;

            string previousName = _name;
            _name = value;
            OnNameChanged?.Invoke(this, previousName);
        }
    }

    public void RemoveTransition(string name)
    {
        if (!possibleTransitions.TryGetValue(name, out StateTransition stateTransition))
            return;

        transitions.Remove(stateTransition);
        possibleTransitions.Remove(name);
    }

    public void AddTransition(string name, StateTransition stateTransition)
    {
        if (transitions.Contains(stateTransition))
            return;

        transitions.Add(stateTransition);
        possibleTransitions.Add(name, stateTransition);
    }

    public void Update()
    {
        if (GetNextState() is IState transitionState)
            OnStateTransitionReady?.InvokeSafe(this, transitionState);
        OnStateUpdate?.InvokeSafe(this);
    }

    public void TransitionTo(IState from) => OnStateTransitionedTo?.InvokeSafe(this, from);
    public void TransitionFrom(IState to) => OnStateTransitionedFrom?.InvokeSafe(this, to);

    public IState? GetNextState()
    {
        foreach (StateTransition stateTransition in transitions)
            if (stateTransition.CanTransition)
                return stateTransition.State;
        return null;
    }
}
