using Syntriax.Engine.Core;

namespace Syntriax.Engine.Systems.Time;

public class StopwatchBehaviour : Behaviour, IStopwatch
{
    public event IReadOnlyStopwatch.StopwatchEventHandler? OnStarted = null;
    public event IReadOnlyStopwatch.StopwatchDeltaEventHandler? OnDelta = null;
    public event IReadOnlyStopwatch.StopwatchEventHandler? OnStopped = null;
    public event IReadOnlyStopwatch.StopwatchEventHandler? OnPaused = null;
    public event IReadOnlyStopwatch.StopwatchEventHandler? OnResumed = null;

    public double Time { get; protected set; } = 0f;
    public TimerState State { get; protected set; } = TimerState.Idle;

    private bool shouldBeTicking = false;
    private bool hasStartedTickingBefore = false;

    public virtual void Start()
    {
        Time = 0f;

        shouldBeTicking = true;
        hasStartedTickingBefore = false;

        if (IsActive)
            StartStopwatch();
    }

    public virtual void Stop()
    {
        if (!shouldBeTicking)
            return;

        shouldBeTicking = false;

        State = TimerState.Stopped;
        OnStopped?.InvokeSafe(this);
    }

    protected override void OnUpdate()
    {
        if (State is not TimerState.Ticking)
            return;

        double delta = Universe.Time.DeltaSpan.TotalSeconds;

        Time += delta;
        OnDelta?.InvokeSafe(this, delta);
    }

    protected override void OnEnteredUniverse(IUniverse universe)
    {
        if (!shouldBeTicking || State is TimerState.Ticking)
            return;

        if (hasStartedTickingBefore)
            Resume();
        else
            StartStopwatch();
    }

    protected override void OnExitedUniverse(IUniverse universe)
    {
        if (!shouldBeTicking || State is not TimerState.Ticking)
            return;

        Pause();
    }

    public virtual void Pause()
    {
        State = TimerState.Paused;
        OnPaused?.InvokeSafe(this);
    }

    public virtual void Resume()
    {
        State = TimerState.Ticking;
        OnResumed?.InvokeSafe(this);
    }

    private void StartStopwatch()
    {
        hasStartedTickingBefore = true;

        State = TimerState.Ticking;
        OnStarted?.InvokeSafe(this);
    }

    protected override void OnFinalize()
    {
        Time = 0f;
        State = TimerState.Idle;
        shouldBeTicking = false;
        hasStartedTickingBefore = false;
    }
}
